# -*- coding: utf-8 -*-

"""Some helperfunctions for clis."""

import sys

from termcolor import cprint

VERBOSITY = 5
DEBUG_PRINT_LEVELS = [
    {"color": "blue", "attrs": ["bold"]},
    {"color": "blue", "attrs": []},
    {"color": "white", "attrs": ["bold"]},
    {"color": "white", "attrs": []},
    {"color": "grey", "attrs": []},
    {"color": "grey", "attrs": ["dark"]},
]


def set_verbosity(verbosity):
    """sets the verbosity level of this script"""
    global VERBOSITY
    VERBOSITY = verbosity


def error(message):
    """prints the message to stderr and exits"""
    cprint(f"[!] {message}", attrs=["bold"], color="red", file=sys.stderr)
    sys.exit(1)


def debug(message, min_debug_level):
    """prints the message to stdout if debug level >= min_debug_level"""
    if min_debug_level <= VERBOSITY:
        cprint(
            f"[i] {message}",
            color=DEBUG_PRINT_LEVELS[min_debug_level]["color"],
            attrs=DEBUG_PRINT_LEVELS[min_debug_level]["attrs"],
        )



"""Fields for models."""

import typing
from abc import abstractmethod, abstractproperty
from collections.abc import Iterable
from datetime import date, datetime

from .errors import ValidationError

if typing.TYPE_CHECKING:
    from enum import Enum


class EmptyParameter:
    """Dummy class for empty parameters."""


# TODO inherit from field?
# TODO differentiate between required for creation and required for instanziation
class Field:
    """The field all fields have to inherit from."""

    def __init__(
        self,
        description: str = "",
        default: typing.Any = EmptyParameter,
        nullable: bool = False,
        choices: "Enum" = None,
        readonly: bool = False,
        validators: typing.List[typing.Callable[[typing.Any], None]] = [],
        is_id_field: bool = False,
    ):
        self.description = description
        self.default = default
        self.nullable = nullable
        self.choices = choices
        self.readonly = readonly
        self.required = default == EmptyParameter
        self.validators = validators
        self.is_id_field = is_id_field

    def clean(self, value: typing.Any, _field_name: typing.Optional[str] = None):
        """Validate and clean the value."""

        if value is None:
            if self.nullable:
                return None
            else:
                raise ValidationError(
                    f"Value must not be None if field '{_field_name}'."
                    if _field_name
                    else "Value must not be None"
                )

        if self.choices:
            choices = {enum.value: enum for enum in self.choices}

            if value in choices:
                value = choices[value]
            else:
                raise ValidationError(
                    f"Invalid value '{value}' for given choices for field '{_field_name}'"
                    if _field_name
                    else f"Invalid value '{value}' for given choices"
                )

        for validator in self.validators:
            validator(value)

        return value

    @abstractproperty
    def python_type(self):
        """Return the corresponding python type."""


class StrField(Field):

    python_type = str

    def clean(self, value, **kwargs):
        if value is not None:
            try:
                value = str(value)
            except (ValueError, TypeError) as err:
                raise ValidationError(err)
            value = value.strip()

        value = super().clean(value, **kwargs)

        return value


class IntField(Field):

    python_type = int

    def clean(self, value, **kwargs):
        if value is not None:
            try:
                value = int(value)
            except (ValueError, TypeError) as err:
                raise ValidationError(err)

        value = super().clean(value, **kwargs)

        return value


class PositiveIntField(IntField):
    def clean(self, value, _field_name: typing.Optional[str] = None):
        value: int = super().clean(value, _field_name=_field_name)

        if value is not None and value < 0:
            raise ValidationError(
                f"Value must not be smaller than zero in field '{_field_name}'."
                if _field_name
                else "Value must not be smaller than zero."
            )

        return value


class BoolField(Field):

    python_type = bool

    def clean(self, value, **kwargs):
        if value is not None:
            try:
                value = bool(value)
            except ValueError as err:
                raise ValidationError(err)

        value = super().clean(value, **kwargs)

        return value


class StrIntField(Field):

    field = None
    python_type = None

    def clean(self, value, **kwargs):
        errs = []
        try:
            field = IntField(
                description=self.description,
                default=self.default,
                nullable=self.nullable,
                choices=self.choices,
                readonly=self.readonly,
                validators=self.validators,
                is_id_field=self.is_id_field,
            )
            value = field.clean(value, **kwargs)
            self.field = field
            self.python_type = int

            return value
        except ValidationError as err:
            errs.append(err)

        try:
            field = StrField(
                description=self.description,
                default=self.default,
                nullable=self.nullable,
                choices=self.choices,
                readonly=self.readonly,
                validators=self.validators,
                is_id_field=self.is_id_field,
            )
            value = field.clean(value, **kwargs)
            self.field = field
            self.python_type = str

            return value
        except ValidationError as err:
            errs.append(err)

        raise ValidationError(errs)


class ListField(Field):
    def __init__(
        self,
        child_fields: typing.List[Field],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.child_fields = child_fields
        # inherit arguments to child fields:

        for child_field in child_fields:
            if child_field.default == EmptyParameter:
                child_field.default = self.default

            if child_field.choices is None:
                child_field.choices = self.choices
            child_field.readonly |= self.readonly
            child_field.validators.extend(self.validators)

    def clean(self, value, _field_name: typing.Optional[str] = None):
        if not isinstance(value, Iterable):
            raise ValidationError(
                f"Value for field '{_field_name}' is not iterable."
                if _field_name
                else "Value is not iterable."
            )

        def clean_subvalue(value, **kwargs):
            validation_errors = []

            for child_field in self.child_fields:
                try:
                    return child_field.clean(value, **kwargs)
                except ValidationError as err:
                    validation_errors.append(err)

            raise ValidationError(validation_errors)

        value = [clean_subvalue(entry, _field_name=_field_name) for entry in value]

        return value


class DateField(Field):

    python_type = date

    def clean(self, value, **kwargs):
        if value is None:
            pass
        elif isinstance(value, date):
            pass
        elif isinstance(value, str):
            value = date.fromisoformat(value)
        elif isinstance(value, int):
            value = date.fromtimestamp(value)
        else:
            ValidationError(value)

        value = super().clean(value, **kwargs)

        return value


class DateTimeField(Field):

    python_type = datetime

    def clean(self, value, **kwargs):
        if value is None:
            pass
        elif isinstance(value, datetime):
            pass
        elif isinstance(value, str):
            value = datetime.fromisoformat(value)
        elif isinstance(value, int):
            value = datetime.fromtimestamp(value)
        else:
            ValidationError(value)

        value = super().clean(value, **kwargs)

        return value



"""The model class all models have to inherit from."""

from __future__ import annotations

import typing

from .fields import Field
from .utils import classproperty


# TODO differentiate between required for creation and required for instanziation
class Model:
    """The model class all models have to inherit from."""

    def __init__(self, **kwargs):
        # evaluate fields
        self._fields = {
            name: value
            for name, value in self.__class__.__dict__.items()
            if isinstance(value, Field)
        }
        self._required_fields = {
            name for name, value in self._fields.items() if value.required
        }
        self._id_field = None

        for name, field in self._fields.items():
            if field.is_id_field:
                if self._id_field:
                    raise ValueError(
                        f"There are multiple id_fields defined for this model {self.__class__}."
                    )
                else:
                    self._id_field = name
        if not self._id_field:
            if "id" in self._fields:
                self._id_field = "id"
            elif "pk" in self._fields:
                self._id_field = "pk"
        # warning no id field defined

        if not kwargs:
            # mode 1: no args, return an empty instance

            return

        # mode 2: all args must be specified

        for name, value in kwargs.items():
            if name in self._fields:
                field = self._fields[name]
                value = field.clean(value)
                setattr(self, value)
            else:
                raise ValueError(f'Invalid argument "{name}".')

        missing_args = set()

        for name, description in self._fields.items():
            value = getattr(self, name)

            if not isinstance(value, Field):
                # field already initialized

                continue
            elif description.required:
                # required field not initialized
                missing_args.add(name)
            else:
                # not required, use default value
                setattr(self, name, description.default)

        if missing_args:
            raise ValueError(f'Missing argument for {", ".join(missing_args)}')

    @classmethod
    def from_api(cls, data: typing.Dict[typing.Any, typing.Any]) -> Model:
        """
        Create an instance from values fetched from the remote API.

        We do not expect to get all values.
        Additional data will be ignored.
        """

        class Unresolved:
            """We did not yet see any value for this field."""

        instance = cls()

        for name, field in instance._fields.items():
            if name in data:
                value = field.clean(data[name], _field_name=name)
                setattr(instance, name, value)
            else:
                setattr(instance, name, Unresolved)

        return instance

    def validate(self):
        """Validate instance for creation."""

        for name, field in self._fields.items():
            value = getattr(self, name)

            if name in self._fields:
                field.validate(value)
            else:
                raise ValueError(f'Invalid argument "{name}".')

    @classproperty
    def resource_name(cls):
        """
        :return: The qualified resource name (e. g. API path) for this class.

        This is a default implementation and can be overridden or set to something static.
        """

        return cls.__name__.rsplit(".", 1)[-1].lower()

    @property
    def id(self):
        """:return: A unique ID."""
        id_field_value = None

        if self._id_field:
            id_field_value = getattr(self, self._id_field, None)

        return id_field_value

    def __str__(self):
        return f"<{self.__class__.__name__} ({self.id})>"



"""A small and simple wrapper for GitPython for GitLab repos."""

import sys

import git

from .script_helpers import error
from .utils import right_replace


def _get_git_repo():
    """:return: a ``git.Repo`` instance."""
    try:
        return git.Repo(search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError as exc:
        error(f"No git repo found: {exc!s}")
        sys.exit(1)


def _get_remote_url():
    """Return the remote from local git repo"""
    repo = _get_git_repo()

    if not repo.remotes:
        error("There are no remotes for this repo. Use --server.")
        sys.exit(1)

    remote = repo.remotes[0]

    if remote.name != "origin":
        for remo in repo.remotes:
            if remo.name == "origin":
                remote = remo

    return remote.url


def get_server_url():
    """
    Return a url to the gitlab server which is the remote of the current repo
    """
    url = _get_remote_url()

    if url.startswith("https://"):
        # https
        url = url.replace("https://", "", 1).split("/")[0]
    elif "@" in url and ":" in url:
        # ssh
        url = "https://{0}/".format(url.split("@")[1].split(":")[0])
    else:
        error(f"The remotes url/protocol is not supported: {url}")

    return url


def get_project_name():
    """
    Return the name of the current repository on the remote gitlab server
    includeing the user/group name (e.g.: 'foogroup/foorepo')
    """
    url = _get_remote_url()

    if url.endswith(".git"):
        url = right_replace(url, ".git", "", 1)

    if "@" in url and ":" in url:
        # ssh
        project = url.split(":", 1)[1]
    elif url.startswith("https://"):
        # https
        project = "/".join(url.replace("https://", "", 1).split("/")[1:])
    else:
        error(f"Unknown repo URL format: {url!s}")
        sys.exit(2)

    return project


def get_shas():
    """Return list of all commits in current repo"""
    repo = _get_git_repo()

    return [ci.hexsha for ci in repo.iter_commits()]


def get_long_commit_sha(hexsha):
    """Return the long hexsha for the commit with sha hexsha"""
    repo = _get_git_repo()

    for commit in repo.iter_commits():
        if commit.hexsha.startswith(hexsha):
            return commit.hexsha

    return None


def get_current_commit():
    """Return the commit hash of the current working copy"""
    repo = _get_git_repo()

    return repo.head.commit.hexsha



"""Some useful functions."""


def right_replace(string: str, old: str, new: str, count: int = -1) -> str:
    """Like ``str.replace`` but go from right to left. Only makes sense for ``count`` > 0."""
    return string[::-1].replace(old[::-1], new[::-1], count)[::-1]



"""A library for django like models and managers."""

from .api_driver import APIDriver, SimpleReadOnlyRESTAPIDriver
from .errors import (
    APIError,
    DoesNotExist,
    OperationalError,
    QueryError,
    ValidationError,
)
from .fields import (
    BoolField,
    DateField,
    DateTimeField,
    Field,
    IntField,
    ListField,
    PositiveIntField,
    StrField,
    StrIntField,
)
from .model import Model



"""The API interface."""

from __future__ import annotations

import logging
import typing
from abc import abstractmethod

from requests import Session

from .errors import APIError, OperationalError
from .model import Model
from .utils import check_single_result

try:
    from hyper.contrib import HTTP20Adapter
except ImportError:
    HTTP20Adapter = None


LOGGER = logging.getLogger()


class APIDriver:
    """
    The abstract class all API implementations must inherit from.

    It follows more or less the CRUD design.
    """

    def __init__(self):
        self._model: typing.Optional[typing.Type[Model]] = None

    def use(self, model: typing.Type[Model]) -> APIDriver:
        """
        Configure the api driver to use a specific model for the next operation(s).
        """
        self._model = model

        return self

    def _check_model_set(self) -> None:
        """Raise an exception if no model is set with ``.use()``."""

        if not self._model:
            raise OperationalError("No model class set. Use `.use()` before.")

    def create(self, instance: Model) -> Model:
        """Create an instance of ``instance`` on the server side."""
        self._check_model_set()
        instance.validate()
        raise NotImplementedError()

    def get(self, **filter) -> typing.Optional[Model]:
        """Get a single instance for a given ``filter``."""
        self._check_model_set()
        raise NotImplementedError()

    def list(
        self,
        filter: typing.Optional[typing.Dict[str, typing.Any]] = None,
        order_by: typing.Optional[typing.List[str]] = None,
    ) -> typing.Iterable[Model]:
        """Get all matching instances for a given ``filter``."""
        self._check_model_set()
        raise NotImplementedError()

    def update(self) -> Model:
        """Update a ressource."""
        self._check_model_set()
        raise NotImplementedError()

    def delete(self) -> None:
        """Delete a ressource."""
        self._check_model_set()
        raise NotImplementedError()

    def __str__(self):
        model_name = self._model.__name__ if self._model else "-"

        return f"<{self.__class__.__name__} ({model_name})>"


class SimpleReadOnlyRESTAPIDriver(APIDriver):
    """A simple read only REST API Driver that may work for some REST APIs."""

    API_PATH = ""

    def __init__(self, host: str):
        super().__init__()
        self.session = Session()
        self.host = host

        if HTTP20Adapter:
            self.session.mount(self.host, HTTP20Adapter())

    @property
    def resource_url(self):
        """:return: The URL to the resource, that is configured using ``.use()``."""
        self._check_model_set()

        return f"{self.host}{self.API_PATH}/{self._model.resource_name}"

    def get(self, **filter):
        matching_instances = self.list(**filter)

        return check_single_result(matching_instances)

    def list(
        self,
        filter: typing.Optional[typing.Dict[str, typing.Any]] = None,
        order_by: typing.Optional[typing.List[str]] = None,
    ) -> typing.Iterable[Model]:
        self._check_model_set()

        response = self.session.get(
            self.resource_url,
            params={
                **(self.create_filter_params(**filter) if filter else {}),
                **(self.create_order_params(*order_by) if order_by else {}),
            },
        )
        try:
            result = response.json()
        except Exception:
            LOGGER.debug("No JSON in response")
            LOGGER.debug(response.text)

        LOGGER.debug(result)

        if not response.ok or result is None:
            raise APIError(result, response)

        if not isinstance(result, list):
            raise OperationalError(
                f"Expected a list as API response but got {type(result)}"
            )

        results = [self._model.from_api(data) for data in result]

        return results

    @abstractmethod
    def create_filter_params(
        self,
        **filter,
    ) -> typing.Dict[str, typing.Union[str, int, bool]]:
        """Create the GET-parameters for a given ``filter``."""

    @abstractmethod
    def create_order_params(
        self,
        *properties,
    ) -> typing.Dict[str, typing.Union[str, int, bool]]:
        """
        Create the GET-parameters for sorting the results.

        :param properties: Property names of the model in this notation `-id`, `+name`, `date_created`.
        """

    def __str__(self):
        model_name = self._model.__name__ if self._model else "-"

        return f"<{self.__class__.__name__} {self.host} ({model_name})>"



"""The API driver for GitLab API."""

import typing

from requests import Session

from gitlabcicli.managerlib import OperationalError, SimpleReadOnlyRESTAPIDriver

from .models import Pipeline, Project

try:
    from hyper.contrib import HTTP20Adapter
except ImportError:
    HTTP20Adapter = None


class GitLabAPIDriver(SimpleReadOnlyRESTAPIDriver):
    """The API driver for GitLab API."""

    API_PATH = "/api/v4"

    def __init__(self, host: str, private_token: typing.Optional[str] = None):
        super().__init__(host)

        if private_token:
            self.session.headers["PRIVATE-TOKEN"] = private_token

    def create_filter_params(
        self,
        **filter,
    ) -> typing.Dict[str, typing.Union[str, int, bool]]:
        self._check_model_set()

        if self._model is Pipeline:
            # TODO

            return filter
        elif self._model is Project:
            if "name" in filter:
                query = {
                    "search": filter.pop("name"),
                }

            if filter:
                raise OperationalError("You only can filter projects by name.")

            return query
        else:
            raise OperationalError(f"Unknown model configured: {self._model}")

    def create_order_params(
        self,
        *properties,
    ) -> typing.Dict[str, typing.Union[str, int, bool]]:
        self._check_model_set()

        if self._model is Pipeline:
            # TODO

            return {}
        elif self._model is Project:
            if len(properties) > 1:
                raise OperationalError("You can only order by one property of Project.")

            prop: str = properties[0]
            sort = "desc" if prop.startswith("-") else "asc"
            prop = prop.lstrip("+-")
            filterable_props = {"id", "name", "created_at", "last_activity_at"}
            if prop not in filterable_props:
                raise OperationalError(
                    f'You can only filter by {", ",join(filterable_props)}'
                )

            return {"order_by": prop, "sort": sort}
        else:
            raise OperationalError(f"Unknown model configured: {self._model}")



# -*- coding: utf-8 -*-

"""Command line interface for GitLab CI to query the current buildstatus."""

__project__ = "gitlabcicli"
__author__ = "sedrubal"
__email__ = "dev@sedrubal.de"
__url__ = f"https://gitlab.com/{__author__}/{__project__}"
__license__ = "CC BY SA 4.0"
__version__ = "0.1"



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

"""Command line interface for GitLab CI."""

import argparse
import os
import sys
import typing
from dataclasses import dataclass
from functools import cached_property
from getpass import getpass
from pathlib import Path

import keyring
from tabulate import tabulate
from termcolor import colored, cprint

import gitlabcicli as info
from gitlabcicli.gitlabapiwrapper import GitLabApiClient
from gitlabcicli.gitwrapper import (
    get_current_commit,
    get_long_commit_sha,
    get_project_name,
    get_server_url,
    get_shas,
)
from gitlabcicli.script_helpers import debug, error, set_verbosity

try:
    import argcomplete
except ImportError:
    argcomplete = None

KNOWN_STATES = {
    "unknown": {
        "color": "yellow",
        "severity": 0,
    },
    "success": {
        "color": "green",
        "severity": 1,
    },
    "pending": {
        "color": "magenta",
        "severity": 2,
    },
    "running": {
        "color": "blue",
        "severity": 3,
    },
    "canceled": {
        "color": "grey",
        "severity": 4,
    },
    "failed": {
        "color": "red",
        "severity": 5,
    },
    "failed (allowed)": {
        "color": "yellow",
        "severity": 2,
    },
    "created": {
        "color": "yellow",
        "severity": 2,
    },
    "skipped": {"color": "blue", "severity": 2},
    "manual": {"color": "red", "severity": 4},
}


@dataclass
class Command:
    """A cmd command."""

    name: str
    aliases: frozenset[str]
    doc: str


COMMANDS = dict[str, Command]()


def register_command(
    *,
    name: typing.Optional[str] = None,
    aliases: set[str] = frozenset[str](),
    doc: typing.Optional[str] = None,
) -> typing.Callable:
    """Decorator to register a command"""

    def inner(func: typing.Callable) -> typing.Callable:
        cmd_name = name or func.__name__.replace("_", "-")
        COMMANDS[cmd_name] = Command(
            name=cmd_name,
            aliases=frozenset(aliases),
            doc=doc or func.__doc__,
        )

        return func

    return inner


KEYRING_SERVICE_NAME = "gitlabcicli"


def sha_completer(prefix, **kwargs):
    """tab completion shas in current git"""

    return (s[:8] for s in get_shas() if s.startswith(prefix))


def gitlabciyml_completer(prefix, **_):
    """tab completion for .gitlab-ci.yml"""
    files = argcomplete.completers.FilesCompleter()(prefix)
    print(files)

    return [
        fn
        for fn in files
        if (os.path.isdir(fn) or ".gitlab-ci.yml".startswith(fn.split("/")[-1]))
    ]


def get_token_from_keyring(server_url: str) -> typing.Optional[str]:
    """Retreive the token for current server from keyring."""

    try:
        return keyring.get_password(KEYRING_SERVICE_NAME, server_url)
    except Exception as exc:
        error(
            "No token found in wallet. Please use --ask-for-token to specify your token."
        )
        error(str(exc))
        sys.exit(500)

    return None


def _color_job_status(status):
    """Return the colored job status as string"""

    return colored(status, attrs=["bold"], color=KNOWN_STATES[status]["color"])


def _color_coverage(coverage):
    """Return the colored coverage as string"""

    if coverage:
        color = "red"

        if coverage > 90:
            color = "green"
        elif coverage > 75:
            color = "yellow"

        return colored(f"{coverage:.1f}%", color=color)
    else:
        debug("No coverage found", min_debug_level=5)

        return ""


def print_job_table(jobs):
    """Print information about the jobs to stdout."""
    pipeline_url = jobs[0]["pipeline"]["web_url"]
    commit_hash = jobs[0]["commit"]["short_id"]
    commit_ref = jobs[0]["ref"]
    commit_title = jobs[0]["commit"]["title"]
    commit_author = jobs[0]["commit"]["author_name"]
    assert all(job["commit"]["short_id"] == commit_hash for job in jobs)
    assert all(job["ref"] == commit_ref for job in jobs)
    assert all(job["commit"]["title"] == commit_title for job in jobs)
    assert all(job["commit"]["author_name"] == commit_author for job in jobs)
    assert all(job["pipeline"]["web_url"] == pipeline_url for job in jobs)
    commit_hash = colored(commit_hash, color="grey")
    commit_ref = colored(commit_ref, color="blue")
    commit_title = colored(commit_title, attrs=["bold"])
    commit_author = colored(commit_author, color="cyan")
    pipeline_url = colored(pipeline_url, attrs=["underline"])
    print(
        f"Commit: {commit_hash} (Ref: {commit_ref}) - {commit_title} - {commit_author}"
    )
    print()
    print(f"Pipeline URL: {pipeline_url}")
    print()

    table_jobs = []
    table_headers = [
        "id",
        "status",
        "stage",
        "name",
        "duration",
        "coverage",
    ]

    def format_duration(duration: float):
        if not duration:
            return ""

        minute = 60
        hour = 60 * minute
        day = 24 * hour
        days, duration = divmod(duration, day)
        hours, duration = divmod(duration, hour)
        minutes, duration = divmod(duration, minute)
        seconds = round(duration)
        ret = ""

        if days:
            ret += f"{days:.0f} days "

        if hours:
            ret += f"{hours:.0f} h "

        if minutes:
            ret += f"{minutes:.0f} min "

        if seconds:
            ret += f"{seconds:.0f} s"

        return ret.strip()

    for job in jobs:
        debug(job["status"], 5)

        if job["status"] not in KNOWN_STATES.keys():
            error(f'[!] Status "{job["status"]}" is unknown!')

            continue
        job_status = job["status"]

        if job_status == "failed" and job.get("allow_failure"):
            job_status = "failed (allowed)"
        table_job = [
            job["id"],
            _color_job_status(job_status),
            job["stage"],
            job["name"],
            format_duration(job.get("duration", 0)),
            _color_coverage(job["coverage"]),
        ]
        table_jobs.append(table_job)
    print(tabulate(table_jobs, headers=table_headers, disable_numparse=True))


class GitLabCiCli(object):
    """A Cli class."""

    def __init__(self, **kwargs):
        self._api_client: typing.Optional[GitLabApiClient] = None
        self._server_url: typing.Optional[str] = kwargs.get("server_url", None)
        self._project_id: typing.Optional[int] = kwargs.get("project_id", None)
        self._project_path: typing.Optional[str] = kwargs.get("project_path", None)
        self._project: typing.Optional[str] = kwargs.get("project", None)
        self._token: typing.Optional[str] = kwargs.get("token", None)
        self._commit_hash: typing.Optional[str] = kwargs.get("commit_hash", None)
        self.args = kwargs

    def stop(self):
        """Shutdown."""

        if self._token:
            keyring.set_password(KEYRING_SERVICE_NAME, self.server_url, self._token)

    @property
    def server_url(self) -> str:
        if not self._server_url:
            debug("Try to get server url from remote of local git repo...", 2)
            self._server_url = get_server_url()
            debug(f"Using server {self._server_url}", 1)

        assert self._server_url

        return self._server_url

    @property
    def api_client(self) -> GitLabApiClient:
        if not self._api_client:
            assert self.token
            self._api_client = GitLabApiClient(
                server_url=self.server_url, token=self.token
            )

        return self._api_client

    @property
    def project_id(self) -> int:
        if self._project_id is None:
            self._project_id = self.api_client.get_project_id(self.project)

            if self._project_id == -1:
                error("Could not determine the project ID.")

        assert self._project_id is not None

        return self._project_id

    @property
    def project_path(self) -> str:
        if not self._project_path:
            self._project_path = self.api_client.get_project_path(self.project_id)

            if not self._project_path:
                error("Could not determine the project path.")

        return self._project_path

    @property
    def commit_hash(self) -> str:
        if not self._commit_hash:
            # ref is better than commit
            debug("Try to get commit from remote of local git repo...", 2)
            self._commit_hash = get_current_commit()
            debug(f"Using commit {self._commit_hash}", 1)

        assert self._commit_hash

        return self._commit_hash

    @property
    def project(self) -> str:
        if not self._project:
            debug("Try to get project from remote of local git repo...", 2)
            self._project = get_project_name()
            debug(f"Using project {self._project}", 1)

        assert self._project

        return self._project

    @cached_property
    def token(self) -> str:
        """The secret API token."""

        if self.args.get("ask_for_token"):
            debug("Try to get token from user input...", 2)
            self._token = getpass(
                f"Please enter the GitLab API Token for {self.server_url}: "
            )
        elif not self._token:
            debug("Try to get token for server from config...", 2)
            self._token = get_token_from_keyring(self.server_url)

        if not self._token:
            error(
                "No token found in wallet or given. Please use --ask-for-token to specify your token."
            )
            sys.exit(501)

        debug(f"Using token {self._token}", 5)

        return self._token

    def get_jobs(self):
        """Return a list of jobs to handle."""

        if "job_ids" in self.args and self.args["job_ids"]:
            jobs = []

            for jid in self.args["job_ids"]:
                job = self.api_client.get_job_for_id(
                    project_id=self.project_id,
                    job_id=jid,
                )

                if job:
                    jobs.append(job)
        else:
            pipeline_id = self.api_client.get_pipeline_for_commit(
                project_id=self.project_id,
                commit_hash=self.commit_hash,
            )

            if not pipeline_id:
                error(f"No pipeline found for commit {self.commit_hash}")
                exit(404)
            jobs = self.api_client.get_pipeline_jobs(
                project_id=self.project_id,
                pipeline_id=pipeline_id,
            )
            jobs = sorted(jobs, key=lambda v: v["id"])

        return jobs

    @register_command(aliases={"status"})
    def show(self):
        """Run gitlabcicli show."""
        jobs = self.get_jobs()

        if not jobs:
            error("No jobs found.")

            return
        print_job_table(jobs)

    @register_command(aliases={"log"})
    def raw(self):
        """run gitlabcicli raw"""
        jobs = self.get_jobs()

        for job in jobs:
            artifacts = job.get("artifacts", [])
            traces = [
                artifact
                for artifact in artifacts
                if artifact.get("file_type") == "trace"
            ]
            job_id = job.get("id")

            if not traces:
                error(
                    f"There is no job log yet for job #{job_id} for project '{self.project}'."
                )

                continue
            elif len(traces) > 1:
                error(
                    f"There are too many job logs for job #{job_id} for project '{self.project}'."
                )

                continue

            trace_file_name = traces.pop().get("filename")
            print()
            cprint(f" === Job output of job #{job['id']} === ", attrs=["bold"])
            print()
            raw_output = self.api_client.get_job_artifact(
                project_path=self.project_path,
                job_id=job["id"],
                file_path=trace_file_name,
            )
            print(raw_output)

            print()

    @register_command(aliases={"action"})
    def do(self):
        """Run gitlabcicli do."""

        for job_id in self.args["job_ids"]:
            response = self.api_client.run_action_on_job(
                action=self.args["job_action"],
                job_id=job_id,
                project_id=self.project_id,
            )
            debug(response, 5)

            if not response:
                error(
                    f"Could not {self.args['job_action']} job #{job_id} for project '{self.project}'."
                )
            else:
                past = {"cancel": "canceled", "retry": "retried", "erase": "erased"}
                cprint(
                    f"Successfully {self.args['job_action']} job #{job_id} for project '{self.project}'",
                    color="green",
                    attrs=["bold"],
                )

                if self.args["job_action"] == "retry":
                    cprint(
                        f"New job id is #{response['id']}",
                        color="green",
                        attrs=["bold"],
                    )

    @register_command(aliases={"validate"})
    def lint(self):
        """Run gitlabcicli lint."""
        gitlabciyml_text = self.args["file"].read()
        debug(f"gitlabci.yml content:\n{gitlabciyml_text}", 5)
        api = self.api_client.validate_ciyml(gitlabciyml_text=gitlabciyml_text)

        if api["status"] == "valid":
            cprint(
                ".gitlab-ci.yml is valid",
                color="green",
                attrs=["bold"],
            )
        else:
            cprint(
                ".gitlab-ci.yml is invalid",
                color="red",
                attrs=["bold"],
            )
            print("List of errors:")

            for error_description in api["errors"]:
                print(f" - {error_description}")

    @register_command(aliases={"web"})
    def open(self):
        """Open the current project in browser."""
        url = self.api_client.get_project_url(self.project_id)
        os.system(f"xdg-open {url}")

    @register_command(aliases={"exec"})
    def run_local(self):
        """Run pipeline on local machine."""
        raise NotImplementedError(
            "run_local is not yet implemented. Contribution is welcome."
        )

    @register_command(aliases={"start"})
    def init(self):
        """Create a .gitlab-ci.yml."""
        # TODO ensure this is in git root
        file_path = Path(".gitlab-ci.yml")

        if file_path.is_file():
            error(f"{file_path} already exists")

        with open(file_path, "w") as file:
            print("---", file=file)
            print("# TODO write here", file=file)
            print("", file=file)
            print("...", file=file)

        print(f"{file_path} created.")


def parse_args():
    """Return the parsed arguments as object (see argparse doc)."""
    # TODO use info from COMMANDS
    parser = argparse.ArgumentParser(description=info.__doc__)
    parser.add_argument(
        "--version", action="version", version="%(prog)s {info.__version__}"
    )
    parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=0,
        help="Be more verbose (up to -vvvvv)",
    )
    parser.add_argument(
        "-t",
        "--ask-for-token",
        action="store_true",
        help="Ask for the GitLab access token using stdin to query the API.",
    )
    parser.add_argument(
        "-s",
        "--server",
        metavar="URL",
        dest="server_url",
        help="The gitlab server url"
        + "(e.g.: https://gitlab.com/, default: parsed from remote)",
    )
    parser.add_argument(
        "-p",
        "--project",
        help="The project name or id (e.g. foo/bar, default: current project)",
    )

    subparsers = parser.add_subparsers(
        title="actions",
        dest="action",
        # required=False, TODO
    )

    # show
    parser_show = subparsers.add_parser(
        "show",
        aliases=COMMANDS["show"].aliases,
        help="Show the current job status",
    )
    parser_show.add_argument(
        "-c",
        "--commit",
        type=get_long_commit_sha,
        metavar="SHA",
        dest="commit_hash",
        help="The commit id to check (default: the current id)",
    ).completer = sha_completer
    parser_show.add_argument(
        "-j",
        "--job",
        metavar="ID",
        dest="job_ids",
        type=int,
        action="append",
        default=[],
        help="The job id to inspect",
    )

    # raw
    parser_raw = subparsers.add_parser(
        "raw",
        aliases=COMMANDS["raw"].aliases,
        help="Get or watch the output of a job",
    )
    parser_raw.add_argument(
        "-f",
        "--follow",
        action="store_true",
        default=False,
        help="Follow running jobs (default: false)",
    )
    parser_raw.add_argument(
        "-c",
        "--commit",
        type=get_long_commit_sha,
        metavar="SHA",
        dest="commit_hash",
        help="The commit id to inspect (default: the current id)",
    ).completer = sha_completer
    parser_raw.add_argument(
        "-j",
        "--job",
        metavar="ID",
        dest="job_ids",
        type=int,
        action="append",
        default=[],
        help="The job id to inspect",
    )

    # do
    parser_do = subparsers.add_parser(
        "do",
        aliases=COMMANDS["do"].aliases,
        help="Run actions on jobs [cancel, retry, erase]",
    )
    parser_do.add_argument(
        "job_action",
        metavar="action",
        choices=("cancel", "retry", "erase"),
        help="What to do with the job [cancel, retry, erase]",
    )
    parser_do.add_argument(
        "-j",
        "--job",
        metavar="ID",
        dest="job_ids",
        type=int,
        action="append",
        required=True,
        help="The job id to inspect",
    )

    # lint
    parser_lint = subparsers.add_parser(
        "lint",
        aliases=COMMANDS["lint"].aliases,
        help="Validate the gitlab-ci.yml",
    )
    parser_lint.add_argument(
        "file",
        type=argparse.FileType("r"),
        default=".gitlab-ci.yml",
        nargs="?",
        help="The gitlab-ci.yml",
    ).completer = gitlabciyml_completer

    # open
    parser_open = subparsers.add_parser(
        "open",
        aliases=COMMANDS["open"].aliases,
        help="Open the current project in your browser.",
    )

    # run local
    parser_run_local = subparsers.add_parser(
        "run-local",
        aliases=COMMANDS["run-local"].aliases,
        help="Run a stage on the local machine.",
    )

    # init
    parser_init = subparsers.add_parser(
        "init",
        aliases=COMMANDS["init"].aliases,
        help="Create a .gitlab-ci.yml",
    )

    if argcomplete:
        argcomplete.autocomplete(parser)

    args = parser.parse_args()

    if args.verbosity > 5:
        args.verbosity = 5
    set_verbosity(args.verbosity)

    if not args.action:
        args.action = "show"

    if args.action not in COMMANDS.keys():
        for command_name, command in COMMANDS.items():
            if args.action in command.aliases:
                args.action = command_name

                break

    return args


def main():
    """The main function"""
    args = parse_args()

    action = COMMANDS[args.action]

    cli = GitLabCiCli(**args.__dict__)
    getattr(cli, action.name)()
    cli.stop()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK

"""Cli to query the buildstate from GitLab CI."""

from gitlabcicli import gitlabcicli

if __name__ == "__main__":
    gitlabcicli.main()



"""Useful functions and more."""

from __future__ import annotations

import typing

from .errors import DoesNotExist, QueryError

if typing.TYPE_CHECKING:
    from .model import Model


class classproperty(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, owner):
        return self.func(owner)


def check_single_result(results: typing.List[Model]) -> Model:
    """
    Check if there is only one result in the results list and return it. Raise exceptions otherwise.
    """

    if len(results) == 1:
        return results[0]

    if len(results) > 1:
        raise QueryError(f"Expected one result but got {len(results)}.")
    raise DoesNotExist()



"""The Project API model and additional things."""

import typing
from datetime import datetime, date
from enum import Enum

from gitlabcicli.managerlib import (
    BoolField,
    DateField,
    DateTimeField,
    IntField,
    ListField,
    Model,
    PositiveIntField,
    StrField,
    StrIntField,
)


class ProjectVisibility(Enum):
    """Possible project visibilities."""

    PRIVATE = "private"
    INTERNAL = "internal"
    PUBLIC = "public"


class Project(Model):
    """The model for a project on GitLab"""

    resource_name = "projects"

    id: int = PositiveIntField()
    description: str = StrField()
    default_branch: str = StrField(default="master", nullable=True)
    visibility = StrField(choices=ProjectVisibility)
    ssh_url_to_repo: str = StrField(readonly=True)
    http_url_to_repo: str = StrField(readonly=True)
    web_url: str = StrField(readonly=True)
    readme_url: str = StrField(readonly=True, nullable=True)
    tag_list: typing.List[str] = ListField([StrField()], readonly=True)
    #  "owner": {
    #      "id": 3,
    #      "name": "Diaspora",
    #      "created_at": "2013-09-30T13:46:02Z"
    #  },
    name: str = StrField()
    name_with_namespace: str = StrField()
    path: str = StrField()
    path_with_namespace: str = StrField()
    issues_enabled: bool = BoolField()
    open_issues_count: int = PositiveIntField(readonly=True)
    merge_requests_enabled: bool = BoolField()
    jobs_enabled: bool = BoolField()
    wiki_enabled: bool = BoolField()
    snippets_enabled: bool = BoolField()
    can_create_merge_request_in: bool = BoolField()
    resolve_outdated_diff_discussions: bool = BoolField()
    container_registry_enabled: bool = BoolField()
    created_at: datetime = DateTimeField(readonly=True)
    last_activity_at: datetime = DateTimeField(readonly=True)
    creator_id: int = PositiveIntField(readonly=True)
    #  "namespace": {
    #      "id": 3,
    #      "name": "Diaspora",
    #      "path": "diaspora",
    #      "kind": "group",
    #      "full_path": "diaspora"
    #  },
    import_status = StrField(readonly=True)
    archived: bool = BoolField(default=False)
    avatar_url: str = StrField(nullable=True, readonly=True)
    shared_runners_enabled: bool = BoolField()
    forks_count: int = PositiveIntField(readonly=True)
    star_count: int = PositiveIntField(readonly=True)
    runners_token: str = StrField()
    ci_default_git_depth: int = PositiveIntField(nullable=True)
    public_jobs: bool = BoolField()
    shared_with_groups: typing.List = ListField([StrIntField()], readonly=True)
    only_allow_merge_if_pipeline_succeeds: bool = BoolField(default=False)
    only_allow_merge_if_all_discussions_are_resolved: bool = BoolField(default=False)
    remove_source_branch_after_merge: bool = BoolField(default=False, nullable=True)
    request_access_enabled: bool = BoolField()
    merge_method: str = StrField(default="merge")
    autoclose_referenced_issues: bool = BoolField(default=True)
    suggestion_commit_message: str = StrField(nullable=True, default=None)
    marked_for_deletion_on: date = DateField(readonly=True, nullable=True, default=None)
    #  "statistics": {
    #      "commit_count": 37,
    #      "storage_size": 1038090,
    #      "repository_size": 1038090,
    #      "wiki_size" : 0,
    #      "lfs_objects_size": 0,
    #      "job_artifacts_size": 0,
    #      "packages_size": 0
    #  },
    #  "_links": {
    #      "self": "http://example.com/api/v4/projects",
    #      "issues": "http://example.com/api/v4/projects/1/issues",
    #      "merge_requests": "http://example.com/api/v4/projects/1/merge_requests",
    #      "repo_branches": "http://example.com/api/v4/projects/1/repository_branches",
    #      "labels": "http://example.com/api/v4/projects/1/labels",
    #      "events": "http://example.com/api/v4/projects/1/events",
    #      "members": "http://example.com/api/v4/projects/1/members"
    #  },



"""The Pipeline API model and additional things."""

import typing
from datetime import datetime
from enum import Enum

from gitlabcicli.managerlib import (
    BoolField,
    DateTimeField,
    IntField,
    Model,
    StrField,
    StrIntField,
)


class Pipeline(Model):
    #  The ID or URL-encoded path of the project owned by the authenticated user
    id: typing.Union[int, str] = StrIntField()
    # The scope of pipelines, one of: running, pending, finished, branches, tags
    scope: str = StrField()
    # The status of pipelines, one of: running, pending, success, failed, canceled, skipped, created, manual
    status: str = StrField()
    # The ref of pipelines
    ref: str = StrField()
    # The SHA of pipelines
    sha: str = StrField()
    # Returns pipelines with invalid configurations
    yaml_errors: bool = BoolField(default=False)
    # The name of the user who triggered pipelines
    name: str = StrField()
    # The username of the user who triggered pipelines
    username: str = StrField()
    # Return pipelines updated after the specified date. Format: ISO 8601 YYYY-MM-DDTHH:MM:SSZ
    updated_after: datetime = DateTimeField()
    # Return pipelines updated before the specified date. Format: ISO 8601 YYYY-MM-DDTHH:MM:SSZ
    updated_before: datetime = DateTimeField()
    # Order pipelines by id, status, ref, updated_at or user_id (default: id)
    order_by: str = StrField()
    # Sort pipelines in asc or desc order (default: desc)
    sort: str = StrField()



"""A small and simple GitLab API wrapper."""

import requests

from gitlabcicli.script_helpers import debug, error

try:
    from hyper.contrib import HTTP20Adapter
except ImportError:
    HTTP20Adapter = None

API4 = "/api/v4/"


class GitLabApiClient:
    """An API Client for the GitLab API."""

    def __init__(self, server_url: str, token: str):
        self.server_url = server_url
        self.session = requests.Session()
        self.session.headers["PRIVATE-TOKEN"] = token

        if HTTP20Adapter:
            self.session.mount(self.server_url, HTTP20Adapter())

    @classmethod
    def _check_response(cls, response, default):
        """
        Check if the api response is a error message and prints the message.

        Return the response json as dict if there is no error.
        """
        try:
            api = response.json()
        except Exception:
            debug("No JSON in response", 4)
            debug(response.text, 5)
            api = None
        debug(api, 5)

        if not response.ok or api is None:
            if isinstance(api, dict) and "message" in api.keys():
                error(api["message"])
            elif isinstance(api, dict) and "error_description" in api.keys():
                error(api["error_description"])
            else:
                error(f"Server Error. Received status {response.status_code}")

            return default

        return api

    def _request(self, url, method="GET", **kwargs):
        """Make a requests."""
        apiurl = f"{self.server_url}{url}"
        debug(f"{method} {apiurl}", 5)

        if method == "GET":
            return self.session.get(apiurl, **kwargs)
        elif method == "POST":
            return self.session.post(apiurl, **kwargs)
        else:
            raise NotImplementedError(f"Method {method} is not implemented.")

    def _get(self, url, **kwargs):
        """Make a GET requests."""

        return self._request(url=url, method="GET", **kwargs)

    def _get_api(self, apiurl, default_return_value=None, **kwargs):
        """Make a GET Request and check returned json."""

        return GitLabApiClient._check_response(
            self._get(url=f"{API4}{apiurl}", **kwargs),
            default=default_return_value,
        )

    def _post_api(self, apiurl, default_return_value=None, **kwargs):
        """Make a POST Request and check returned json."""

        return GitLabApiClient._check_response(
            self._request(url=f"{API4}{apiurl}", method="POST", **kwargs),
            default=default_return_value,
        )

    def get_pipeline_for_commit(self, project_id, commit_hash):
        """Return the pipeline id for the commit in project on server"""
        commit = self._get_api(
            apiurl=f"projects/{project_id}/repository/commits/{commit_hash}/",
            default_return_value={},
        )
        try:
            pipeline_id = (
                commit["last_pipeline"]["id"] if commit["last_pipeline"] else None
            )
            debug(f"Pipeline ID for commit {commit_hash} is {pipeline_id}", 5)
        except KeyError:
            pipeline_id = None

        return pipeline_id

    def get_pipeline_jobs(self, project_id, pipeline_id):
        """Return a list of jobs for a pipeline."""
        jobs = self._get_api(
            apiurl=f"projects/{project_id}/pipelines/{pipeline_id}/jobs/",
            default_return_value=[],
        )

        return jobs

    def get_job_for_id(self, project_id, job_id):
        """Return the job for the given job id"""
        job = self._get_api(
            apiurl=f"projects/{project_id}/jobs/{job_id}",
            default_return_value={},
        )

        return job

    def get_job_artifact(self, project_path: str, job_id, file_path: str):
        """Return the raw job output."""
        response = self._get_api(
            apiurl=f"/projects/{project_path}/jobs/{job_id}/artifacts/{file_path}"
        )
        debug(response.text, 5)

        if not response.ok:
            if response.status_code == 404:
                error(f"No log for job {job_id} found.")

                return ""
            else:
                error(
                    f"Could not get raw job log. Received status code {response.status_code}"
                )

        return response.text

    def get_project_path(self, project_id):
        """Return the full path with namespace for the project with id `project_id`."""

        if not str(project_id).isnumeric:
            debug("interpreting project_id as path.", 4)

            return str(project_id)

        project = self._get_api(
            apiurl=f"projects/{project_id}/",
            default_return_value={},
        )

        return project.get("path_with_namespace", None)

    def get_project_id(self, project: str):
        """Return the id for the project on server"""

        if project.isnumeric():
            debug("interpreting project argument as id.", 3)

            return int(project)

        query = project.split("/", 2)[-1]
        api = self._get_api(
            apiurl=f"projects/?search={query}",
            default_return_value=[],
            params={"order_by": "last_activity_at", "sort": "desc"},
        )

        for apiproj in api:
            if apiproj["path_with_namespace"] == project:
                pid = apiproj["id"]
                debug(f'Project ID for "{project}" is {pid}', 4)

                return pid
        error(f'Project "{project}" not found on server "{self.server_url}"!')

        return -1

    def get_project_url(self, project_id: int):
        """Return the http url for the project with id `project_id`."""
        project = self._get_api(
            apiurl=f"projects/{project_id}/",
            default_return_value={},
        )

        return project.get("web_url", None)

    def run_action_on_job(self, action, project_id, job_id):
        """Run action (cancel, retry, erase) on job"""

        return self._post_api(
            apiurl=f"projects/{project_id}/jobs/{job_id}/{action}",
            default_return_value={},
        )

    def validate_ciyml(self, gitlabciyml_text):
        """Return:
        valid:
            {"status": "valid", "errors": []}
        invalid:
            {"status": "invalid", "errors": ["description 1", "description 2"]}
        error:
            {"error": "description"}
        """
        api = self._post_api(
            apiurl="/ci/lint",
            default_return_value={"error": "request failed"},
            data={"content": gitlabciyml_text},
        )

        if "error" in api.keys():
            error(api["error"])
        elif (
            "status" not in api.keys()
            or "errors" not in api.keys()
            or api["status"] not in ["valid", "invalid"]
        ):
            debug(api, 4)
            error("Invalid response from api")
        else:
            return api



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gitlabcicli',
 'gitlabcicli.api',
 'gitlabcicli.api.models',
 'gitlabcicli.managerlib']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.11,<4.0.0',
 'argcomplete>=1.12.2,<2.0.0',
 'dbus-python>=1.2.16,<2.0.0',
 'keyring>=21.5.0,<22.0.0',
 'requests>=2.25.0,<3.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'termcolor>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['gitlabcicli = gitlabcicli.gitlabcicli:main']}

setup_kwargs = {
    'name': 'gitlabcicli',
    'version': '0.2.2',
    'description': 'Command line interface for GitLab CI',
    'long_description': None,
    'author': 'sedrubal',
    'author_email': 'dev@sedrubal.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/sedrubal/gitlabcicli.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""The errors that are raised by the lib."""


class OperationalError(Exception):
    """A server side operation could not be executed."""


class ValidationError(ValueError):
    """Raised, when a value could not be validated."""


class QueryError(Exception):
    """A error was caused by a specific query."""


class DoesNotExist(QueryError):
    """A ressource matching the query does not exist."""


class APIError(Exception):
    """The API returned an error."""
